@@ -2,7 +2,7 @@
 
 ACLOCAL="aclocal -I m4 -I config"
 AUTOHEADER="autoheader"
-AUTOMAKE="automake --add-missing --copy --foreign"
+AUTOMAKE="automake --add-missing --force-missing --copy --foreign"
 AUTOCONF="autoconf"
 
 # Discover what version of automake we are using.
@@ -16,12 +16,12 @@ test "$automake_version" != "" && {
 }
 
 case $automake_version in
-    *' '1.[8-9]*|*' '1.1[01]*)
+    *' '1.[8-9]*|*' '1.1[01234]*)
 	;;
     *)
 
 	echo "This automake version is not supported by LyX."
-	echo "LyX only supports automake 1.8 to 1.11."
+	echo "LyX only supports automake 1.8 to 1.14."
 	exit 1
 	;;
 esac
@@ -32,17 +32,17 @@ autoversion=`$AUTOCONF --version 2>/dev/null | head -n 1`
 test "$autoversion" != "" && {
     echo "Using $autoversion"
 } || {
-    echo "LyX requires autoconf >= 2.59c"
+    echo "LyX requires autoconf >= 2.60"
     exit 1
 }
 
 
 case $autoversion in
-    *' '2.59[cd]|*' '2.60[ab]|*' '2.6[0-8])
+    *' '2.60[ab]|*' '2.6[0-9])
 	;;
     *)
 	echo "This autoconf version is not supported by LyX."
-	echo "LyX only supports autoconf 2.59c-2.68."
+	echo "LyX only supports autoconf 2.60-2.69."
 	exit 1
 	;;
 esac
@@ -85,8 +85,11 @@ else
 fi
 
 echo "Building po/POTFILES.in..."
-make -s -f po/Rules-lyx srcdir=po top_srcdir=. po/POTFILES.in
+if (! make -s -f po/Rules-lyx srcdir=po top_srcdir=. po/POTFILES.in ); then
+	echo "Building po/POTFILES.in failed -- aborting"
+	exit 1
+fi
 
 echo
-echo 'run "./configure ; make"'
+echo 'run "./configure && make"'
 echo
