@@ -1,4 +1,3 @@
-#! /usr/bin/env python
 # -*- coding: utf-8 -*-
 
 # file lyxpreview2bitmap.py
@@ -16,6 +15,7 @@
 # png or ppm image files, one per previewed snippet.
 
 # Pre-requisites:
+# * python 2.4 or later (subprocess module);
 # * A latex executable;
 # * preview.sty;
 # * dvipng;
@@ -29,16 +29,28 @@
 # CTAN/support/preview-latex/
 
 # Example usage:
-# lyxpreview2bitmap.py png 0lyxpreview.tex 128 000000 faf0e6
+# lyxpreview2bitmap.py --bg=faf0e6 0lyxpreview.tex
 
-# This script takes six arguments:
-# FORMAT:   The desired output format. Either 'png' or 'ppm'.
-# TEXFILE:  the name of the .tex file to be converted.
-# DPI:      a scale factor, used to ascertain the resolution of the
-#           generated image which is then passed to gs.
-# FG_COLOR: the foreground color as a hexadecimal string, eg '000000'.
-# BG_COLOR: the background color as a hexadecimal string, eg 'faf0e6'.
-# CONVERTER: the converter (optional). Default is latex.
+# This script takes one obligatory argument:
+#
+#   <input file>:  The name of the .tex file to be converted.
+#
+# and these optional arguments:
+#
+#   --png, --ppm:  The desired output format. Either 'png' or 'ppm'.
+#   --dpi=<res>:   A scale factor, used to ascertain the resolution of the
+#                  generated image which is then passed to gs.
+#   --fg=<color>:  The foreground color as a hexadecimal string, eg '000000'.
+#   --bg=<color>:  The background color as a hexadecimal string, eg 'faf0e6'.
+#   --latex=<exe>: The converter for latex files. Default is latex.
+#   --bibtex=<exe>: The converter for bibtex files. Default is bibtex.
+#   --lilypond:    Preprocess through lilypond-book. Default is false.
+#   --lilypond-book=<exe>:
+#                  The converter for lytex files. Default is lilypond-book.
+#
+#   -d, --debug    Show the output from external commands.
+#   -h, --help     Show an help screen and exit.
+#   -v, --verbose  Show progress messages.
 
 # Decomposing TEXFILE's name as DIR/BASE.tex, this script will,
 # if executed successfully, leave in DIR:
@@ -50,8 +62,8 @@
 # What does this script do?
 # 1) Call latex/pdflatex/xelatex/whatever (CONVERTER parameter)
 # 2) If the output is a PDF fallback to legacy
-# 3) Otherwise check each page of the DVI (with dv2dt) looking for 
-#    PostScript literals, not well supported by dvipng. Pages 
+# 3) Otherwise check each page of the DVI (with dv2dt) looking for
+#    PostScript literals, not well supported by dvipng. Pages
 #    containing them are passed to the legacy method in a new LaTeX file.
 # 4) Call dvipng on the pages without PS literals
 # 5) Join metrics info coming from both methods (legacy and dvipng)
@@ -63,27 +75,38 @@
 # Moreover dvipng can't work with PDF files, so, if the CONVERTER
 # paramter is pdflatex we have to fallback to legacy route (step 2).
 
-import glob, os, re, string, sys
+import getopt, glob, os, re, shutil, string, sys
 
-from legacy_lyxpreview2ppm import legacy_conversion, \
-     legacy_conversion_step2, legacy_extract_metrics_info, filter_pages
-     
-from lyxpreview_tools import copyfileobj, error, find_exe, \
-     find_exe_or_terminate, make_texcolor, mkstemp, run_command, warning, \
-     write_metrics_info, join_metrics_and_rename
+from legacy_lyxpreview2ppm import legacy_conversion_step1
 
+from lyxpreview_tools import bibtex_commands, copyfileobj, error, \
+     filter_pages, find_exe, find_exe_or_terminate, join_metrics_and_rename, \
+     latex_commands, latex_file_re, make_texcolor, mkstemp, pdflatex_commands, \
+     progress, run_command, run_latex, run_tex, warning, write_metrics_info
 
-# Pre-compiled regular expressions.
-latex_file_re = re.compile("\.tex$")
-
-# PATH environment variable
-path  = string.split(os.environ["PATH"], os.pathsep)
 
 def usage(prog_name):
-    return "Usage: %s <format> <latex file> <dpi> <fg color> <bg color>\n"\
-           "\twhere the colors are hexadecimal strings, eg 'faf0e6'"\
-           % prog_name
-    
+    msg = """
+Usage: %s <options> <input file>
+
+Options:
+  --dpi=<res>:   Resolution per inch (default: 128)
+  --png, --ppm:  Select the output format (default: png)
+  --fg=<color>:  Foreground color (default: black, ie '000000')
+  --bg=<color>:  Background color (default: white, ie 'ffffff')
+  --latex=<exe>: Specify the executable for latex (default: latex)
+  --bibtex=<exe>: Specify the executable for bibtex (default: bibtex)
+  --lilypond:    Preprocess through lilypond-book (default: false)
+  --lilypond-book=<exe>:
+                 The executable for lilypond-book (default: lilypond-book)
+
+  -d, --debug:   Show the output from external commands
+  -h, --help:    Show this help screen and exit
+  -v, --verbose: Show progress messages
+
+The colors are hexadecimal strings, eg 'faf0e6'."""
+    return msg % prog_name
+
 # Returns a list of tuples containing page number and ascent fraction
 # extracted from dvipng output.
 # Use write_metrics_info to create the .metrics file with this info
@@ -127,46 +150,10 @@ def extract_metrics_info(dvipng_stdout):
 
     if success == 0:
         error("Failed to extract metrics info from dvipng")
-    
+
     return results
 
 
-def color_pdf(latex_file, bg_color, fg_color):
-    use_preview_pdf_re = re.compile("(\s*\\\\usepackage\[[^]]+)((pdftex|xetex)\]{preview})")
-
-    tmp = mkstemp()
-    
-    fg = ""
-    if fg_color != "0.000000,0.000000,0.000000":
-        fg = '  \\AtBeginDocument{\\let\\oldpreview\\preview\\renewcommand\\preview{\\oldpreview\\color[rgb]{%s}}}\n' % (fg_color)
-    
-    success = 0
-    try:
-        for line in open(latex_file, 'r').readlines():
-            match = use_preview_pdf_re.match(line)
-            if match == None:
-                tmp.write(line)
-                continue
-            success = 1
-            tmp.write("  \\usepackage{color}\n" \
-                  "  \\pagecolor[rgb]{%s}\n" \
-                  "%s" \
-                  "%s\n" \
-                  % (bg_color, fg, match.group()))
-            continue
-
-    except:
-        # Unable to open the file, but do nothing here because
-        # the calling function will act on the value of 'success'.
-        warning('Warning in color_pdf! Unable to open "%s"' % latex_file)
-        warning(`sys.exc_type` + ',' + `sys.exc_value`)
-
-    if success:
-        copyfileobj(tmp, open(latex_file,"wb"), 1)
-
-    return success
-
-
 def fix_latex_file(latex_file):
     documentclass_re = re.compile("(\\\\documentclass\[)(1[012]pt,?)(.+)")
 
@@ -185,7 +172,7 @@ def fix_latex_file(latex_file):
     if changed:
         copyfileobj(tmp, open(latex_file,"wb"), 1)
 
-    return
+    return changed
 
 
 def convert_to_ppm_format(pngtopnm, basename):
@@ -196,7 +183,7 @@ def convert_to_ppm_format(pngtopnm, basename):
 
         p2p_cmd = '%s "%s"' % (pngtopnm, png_file)
         p2p_status, p2p_stdout = run_command(p2p_cmd)
-        if p2p_status != None:
+        if p2p_status:
             error("Unable to convert %s to ppm format" % png_file)
 
         ppm = open(ppm_file, 'w')
@@ -212,12 +199,12 @@ def find_ps_pages(dvi_file):
     # FIXME: try with pdflatex
     if not os.path.isfile(dvi_file):
         error("No DVI output.")
-        
+
     # Check for PostScript specials in the dvi, badly supported by dvipng
     # This is required for correct rendering of PSTricks and TikZ
-    dv2dt = find_exe_or_terminate(["dv2dt"], path)
+    dv2dt = find_exe_or_terminate(["dv2dt"])
     dv2dt_call = '%s "%s"' % (dv2dt, dvi_file)
- 
+
     # The output from dv2dt goes to stdout
     dv2dt_status, dv2dt_output = run_command(dv2dt_call)
     psliteral_re = re.compile("^special[1-4] [0-9]+ '(\"|ps:)")
@@ -243,7 +230,7 @@ def find_ps_pages(dvi_file):
         if psliteral_re.match(line) != None:
             # Literal PostScript special detected!
             page_has_ps = True
-    
+
     # Create the -pp parameter for dvipng
     pages_parameter = ""
     if len(ps_pages) > 0 and len(ps_pages) < page_index:
@@ -286,61 +273,144 @@ def find_ps_pages(dvi_file):
     return (ps_pages, page_index, pages_parameter)
 
 def main(argv):
+    # Set defaults.
+    dpi = 128
+    fg_color = "000000"
+    bg_color = "ffffff"
+    bibtex = None
+    latex = None
+    lilypond = False
+    lilypond_book = None
+    output_format = "png"
+    script_name = argv[0]
+
     # Parse and manipulate the command line arguments.
-    if len(argv) != 6 and len(argv) != 7:
-        error(usage(argv[0]))
+    try:
+        (opts, args) = getopt.gnu_getopt(argv[1:], "dhv", ["bibtex=", "bg=",
+            "debug", "dpi=", "fg=", "help", "latex=", "lilypond",
+            "lilypond-book=", "png", "ppm", "verbose"])
+    except getopt.GetoptError, err:
+        error("%s\n%s" % (err, usage(script_name)))
 
-    output_format = string.lower(argv[1])
+    opts.reverse()
+    for opt, val in opts:
+        if opt in ("-h", "--help"):
+            print usage(script_name)
+            sys.exit(0)
+        elif opt == "--bibtex":
+            bibtex = [val]
+        elif opt == "--bg":
+            bg_color = val
+        elif opt in ("-d", "--debug"):
+            import lyxpreview_tools
+            lyxpreview_tools.debug = True
+        elif opt == "--dpi":
+            try:
+                dpi = string.atoi(val)
+            except:
+                error("Cannot convert %s to an integer value" % val)
+        elif opt == "--fg":
+            fg_color = val
+        elif opt == "--latex":
+            latex = [val]
+        elif opt == "--lilypond":
+            lilypond = True
+        elif opt == "--lilypond-book":
+            lilypond_book = [val]
+        elif opt in ("--png", "--ppm"):
+            output_format = opt[2:]
+        elif opt in ("-v", "--verbose"):
+            import lyxpreview_tools
+            lyxpreview_tools.verbose = True
 
-    dir, latex_file = os.path.split(argv[2])
+    # Determine input file
+    if len(args) != 1:
+        err = "A single input file is required, %s given" % (len(args) or "none")
+        error("%s\n%s" % (err, usage(script_name)))
+
+    input_path = args[0]
+    dir, latex_file = os.path.split(input_path)
+
+    # Echo the settings
+    progress("Starting %s..." % script_name)
+    progress("Output format: %s" % output_format)
+    progress("Foreground color: %s" % fg_color)
+    progress("Background color: %s" % bg_color)
+    progress("Resolution (dpi): %s" % dpi)
+    progress("File to process: %s" % input_path)
+
+    # Check for the input file
+    if not os.path.exists(input_path):
+        error('File "%s" not found.' % input_path)
     if len(dir) != 0:
         os.chdir(dir)
 
-    dpi = string.atoi(argv[3])
-    fg_color = make_texcolor(argv[4], False)
-    bg_color = make_texcolor(argv[5], False)
-
-    fg_color_gr = make_texcolor(argv[4], True)
-    bg_color_gr = make_texcolor(argv[5], True)
+    fg_color_dvipng = make_texcolor(fg_color, False)
+    bg_color_dvipng = make_texcolor(bg_color, False)
 
     # External programs used by the script.
-    if len(argv) == 7:
-        latex = argv[6]
-    else:
-        latex = find_exe_or_terminate(["latex", "pplatex", "platex", "latex2e"], path)
+    latex = find_exe_or_terminate(latex or latex_commands)
+    bibtex = find_exe(bibtex or bibtex_commands)
+    if lilypond:
+        lilypond_book = find_exe_or_terminate(lilypond_book or
+            ["lilypond-book --safe"])
+
+    # These flavors of latex are known to produce pdf output
+    pdf_output = latex in pdflatex_commands
+
+    progress("Latex command: %s" % latex)
+    progress("Latex produces pdf output: %s" % pdf_output)
+    progress("Bibtex command: %s" % bibtex)
+    progress("Lilypond-book command: %s" % lilypond_book)
+    progress("Preprocess through lilypond-book: %s" % lilypond)
+    progress("Altering the latex file for font size and colors")
 
     # Omit font size specification in latex file.
     fix_latex_file(latex_file)
 
+    if lilypond:
+        progress("Preprocess the latex file through %s" % lilypond_book)
+        if pdf_output:
+            lilypond_book += " --pdf"
+        lilypond_book += " --latex-program=%s" % latex.split()[0]
+
+        # Make a copy of the latex file
+        lytex_file = latex_file_re.sub(".lytex", latex_file)
+        shutil.copyfile(latex_file, lytex_file)
+
+        # Preprocess the latex file through lilypond-book.
+        lytex_status, lytex_stdout = run_tex(lilypond_book, lytex_file)
+
+    if pdf_output:
+        progress("Using the legacy conversion method (PDF support)")
+        return legacy_conversion_step1(latex_file, dpi, output_format, fg_color,
+            bg_color, latex, pdf_output)
+
     # This can go once dvipng becomes widespread.
-    dvipng = find_exe(["dvipng"], path)
+    dvipng = find_exe(["dvipng"])
     if dvipng == None:
-        # The data is input to legacy_conversion in as similar
-        # as possible a manner to that input to the code used in
-        # LyX 1.3.x.
-        vec = [ argv[0], argv[2], argv[3], argv[1], argv[4], argv[5], latex ]
-        return legacy_conversion(vec)
+        progress("Using the legacy conversion method (dvipng not found)")
+        return legacy_conversion_step1(latex_file, dpi, output_format, fg_color,
+            bg_color, latex, pdf_output)
+
+    dv2dt = find_exe(["dv2dt"])
+    if dv2dt == None:
+        progress("Using the legacy conversion method (dv2dt not found)")
+        return legacy_conversion_step1(latex_file, dpi, output_format, fg_color,
+            bg_color, latex, pdf_output)
 
     pngtopnm = ""
     if output_format == "ppm":
-        pngtopnm = find_exe_or_terminate(["pngtopnm"], path)
-
-    # Move color information for PDF into the latex file.
-    if not color_pdf(latex_file, bg_color_gr, fg_color_gr):
-        error("Unable to move color info into the latex file")
+        pngtopnm = find_exe(["pngtopnm"])
+        if pngtopnm == None:
+            progress("Using the legacy conversion method (pngtopnm not found)")
+            return legacy_conversion_step1(latex_file, dpi, output_format,
+                fg_color, bg_color, latex, pdf_output)
 
     # Compile the latex file.
-    latex_call = '%s "%s"' % (latex, latex_file)
-
-    latex_status, latex_stdout = run_command(latex_call)
-    if latex_status != None:
-        warning("%s had problems compiling %s" \
-              % (os.path.basename(latex), latex_file))
-
-    if latex == "xelatex":
-        warning("Using XeTeX")
-        # FIXME: skip unnecessary dvips trial in legacy_conversion_step2
-        return legacy_conversion_step2(latex_file, dpi, output_format)
+    latex_status, latex_stdout = run_latex(latex, latex_file, bibtex)
+    if latex_status:
+      return (latex_status, [])
 
     # The dvi output file name
     dvi_file = latex_file_re.sub(".dvi", latex_file)
@@ -350,9 +420,11 @@ def main(argv):
         # No DVI, is there a PDF?
         pdf_file = latex_file_re.sub(".pdf", latex_file)
         if os.path.isfile(pdf_file):
-            warning("%s produced a PDF output, fallback to legacy." % \
-                (os.path.basename(latex)))
-            return legacy_conversion_step2(latex_file, dpi, output_format)
+            progress("%s produced a PDF output, fallback to legacy." \
+                % (os.path.basename(latex)))
+            progress("Using the legacy conversion method (PDF support)")
+            return legacy_conversion_step1(latex_file, dpi, output_format,
+                fg_color, bg_color, latex, True)
         else:
             error("No DVI or PDF output. %s failed." \
                 % (os.path.basename(latex)))
@@ -362,22 +434,24 @@ def main(argv):
     # page_count: total number of pages
     # pages_parameter: parameter for dvipng to exclude pages with PostScript
     (ps_pages, page_count, pages_parameter) = find_ps_pages(dvi_file)
-    
+
     # If all pages need PostScript, directly use the legacy method.
     if len(ps_pages) == page_count:
-        vec = [argv[0], argv[2], argv[3], argv[1], argv[4], argv[5], latex]
-        return legacy_conversion(vec)
+        progress("Using the legacy conversion method (PostScript support)")
+        return legacy_conversion_step1(latex_file, dpi, output_format, fg_color,
+            bg_color, latex, pdf_output)
 
     # Run the dvi file through dvipng.
     dvipng_call = '%s -Ttight -depth -height -D %d -fg "%s" -bg "%s" %s "%s"' \
-                  % (dvipng, dpi, fg_color, bg_color, pages_parameter, dvi_file)
+        % (dvipng, dpi, fg_color_dvipng, bg_color_dvipng, pages_parameter, dvi_file)
     dvipng_status, dvipng_stdout = run_command(dvipng_call)
 
-    if dvipng_status != None:
+    if dvipng_status:
         warning("%s failed to generate images from %s... fallback to legacy method" \
               % (os.path.basename(dvipng), dvi_file))
-        # FIXME: skip unnecessary dvips trial in legacy_conversion_step2
-        return legacy_conversion_step2(latex_file, dpi, output_format)
+        progress("Using the legacy conversion method (dvipng failed)")
+        return legacy_conversion_step1(latex_file, dpi, output_format, fg_color,
+            bg_color, latex, pdf_output)
 
     # Extract metrics info from dvipng_stdout.
     metrics_file = latex_file_re.sub(".metrics", latex_file)
@@ -391,16 +465,17 @@ def main(argv):
         filter_pages(latex_file, legacy_latex_file, ps_pages)
 
         # Pass the new LaTeX file to the legacy method
-        vec = [ argv[0], latex_file_re.sub("_legacy.tex", argv[2]), \
-                argv[3], argv[1], argv[4], argv[5], latex ]
-        legacy_metrics = legacy_conversion(vec, True)[1]
-        
+        progress("Pages %s include postscript specials" % ps_pages)
+        progress("Using the legacy conversion method (PostScript support)")
+        legacy_status, legacy_metrics = legacy_conversion_step1(legacy_latex_file,
+            dpi, output_format, fg_color, bg_color, latex, pdf_output, True)
+
         # Now we need to mix metrics data from dvipng and the legacy method
         original_bitmap = latex_file_re.sub("%d." + output_format, legacy_latex_file)
         destination_bitmap = latex_file_re.sub("%d." + output_format, latex_file)
-        
+
         # Join metrics from dvipng and legacy, and rename legacy bitmaps
-        join_metrics_and_rename(dvipng_metrics, legacy_metrics, ps_pages, 
+        join_metrics_and_rename(dvipng_metrics, legacy_metrics, ps_pages,
             original_bitmap, destination_bitmap)
 
     # Convert images to ppm format if necessary.
@@ -409,8 +484,8 @@ def main(argv):
 
     # Actually create the .metrics file
     write_metrics_info(dvipng_metrics, metrics_file)
-    
+
     return (0, dvipng_metrics)
 
 if __name__ == "__main__":
-    exit(main(sys.argv)[0])
+    sys.exit(main(sys.argv)[0])
