@@ -45,13 +45,13 @@
 #include "SpellChecker.h"
 #include "TextClass.h"
 #include "TextMetrics.h"
-#include "VSpace.h"
 #include "WordLangTuple.h"
 
 #include "frontends/Application.h"
 #include "frontends/Clipboard.h"
 #include "frontends/Selection.h"
 
+#include "insets/InsetArgument.h"
 #include "insets/InsetCollapsable.h"
 #include "insets/InsetCommand.h"
 #include "insets/InsetExternal.h"
@@ -59,6 +59,7 @@
 #include "insets/InsetFloatList.h"
 #include "insets/InsetGraphics.h"
 #include "insets/InsetGraphicsParams.h"
+#include "insets/InsetIPAMacro.h"
 #include "insets/InsetNewline.h"
 #include "insets/InsetQuotes.h"
 #include "insets/InsetSpecialChar.h"
@@ -68,9 +69,11 @@
 #include "support/convert.h"
 #include "support/debug.h"
 #include "support/gettext.h"
+#include "support/lassert.h"
 #include "support/lstrings.h"
 #include "support/lyxtime.h"
 #include "support/os.h"
+#include "support/regex.h"
 
 #include "mathed/InsetMathHull.h"
 #include "mathed/MathMacroTemplate.h"
@@ -93,6 +96,7 @@ using cap::pasteClipboardGraphics;
 using cap::replaceSelection;
 using cap::grabAndEraseSelection;
 using cap::selClearOrDel;
+using cap::pasteSimpleText;
 
 // globals...
 static Font freefont(ignore_font, ignore_language);
@@ -145,7 +149,7 @@ static void mathDispatch(Cursor & cur, FuncRequest const & cmd, bool display)
 #endif
 		cur.insert(new InsetMathHull(cur.buffer(), hullSimple));
 #ifdef ENABLE_ASSERTIONS
-		LASSERT(old_pos == cur.pos(), /**/);
+		LATTEST(old_pos == cur.pos());
 #endif
 		cur.nextInset()->edit(cur, true);
 		// don't do that also for LFUN_MATH_MODE
@@ -226,6 +230,15 @@ static void specialChar(Cursor & cur, InsetSpecialChar::Kind kind)
 }
 
 
+static void ipaChar(Cursor & cur, InsetIPAChar::Kind kind)
+{
+	cur.recordUndo();
+	cap::replaceSelection(cur);
+	cur.insert(new InsetIPAChar(kind));
+	cur.posForward();
+}
+
+
 static bool doInsertInset(Cursor & cur, Text * text,
 	FuncRequest const & cmd, bool edit, bool pastesel)
 {
@@ -283,7 +296,14 @@ static bool doInsertInset(Cursor & cur, Text * text,
 			// Merge multiple paragraphs -- hack
 			while (cur.lastpit() > 0)
 				mergeParagraph(bparams, cur.text()->paragraphs(), 0);
+			if (cmd.action() == LFUN_FLEX_INSERT)
+				return true;
+			Cursor old = cur;
 			cur.leaveInset(*inset);
+			if (cmd.action() == LFUN_PREVIEW_INSERT
+			    || cmd.action() == LFUN_IPA_INSERT)
+				// trigger preview
+				notifyCursorLeavesOrEnters(old, cur);
 		}
 	} else {
 		cur.leaveInset(*inset);
@@ -327,7 +347,7 @@ static void outline(OutlineOp mode, Cursor & cur)
 
 	DocumentClass const & tc = buf.params().documentClass();
 
-	int const thistoclevel = start->layout().toclevel;
+	int const thistoclevel = buf.text().getTocLevel(distance(bgn, start));
 	int toclevel;
 
 	// Move out (down) from this section header
@@ -336,7 +356,7 @@ static void outline(OutlineOp mode, Cursor & cur)
 
 	// Seek the one (on same level) below
 	for (; finish != end; ++finish) {
-		toclevel = finish->layout().toclevel;
+		toclevel = buf.text().getTocLevel(distance(bgn, finish));
 		if (toclevel != Layout::NOT_IN_TOC && toclevel <= thistoclevel)
 			break;
 	}
@@ -353,7 +373,7 @@ static void outline(OutlineOp mode, Cursor & cur)
 			// Search previous same-level header above
 			do {
 				--dest;
-				toclevel = dest->layout().toclevel;
+				toclevel = buf.text().getTocLevel(distance(bgn, dest));
 			} while(dest != bgn
 				&& (toclevel == Layout::NOT_IN_TOC
 				    || toclevel > thistoclevel));
@@ -376,7 +396,7 @@ static void outline(OutlineOp mode, Cursor & cur)
 			ParagraphList::iterator dest = boost::next(finish, 1);
 			// Go further down to find header to insert in front of:
 			for (; dest != end; ++dest) {
-				toclevel = dest->layout().toclevel;
+				toclevel = buf.text().getTocLevel(distance(bgn, dest));
 				if (toclevel != Layout::NOT_IN_TOC
 				      && toclevel <= thistoclevel)
 					break;
@@ -393,7 +413,7 @@ static void outline(OutlineOp mode, Cursor & cur)
 			pit_type const len = distance(start, finish);
 			buf.undo().recordUndo(cur, ATOMIC_UNDO, pit, pit + len - 1);
 			for (; start != finish; ++start) {
-				toclevel = start->layout().toclevel;
+				toclevel = buf.text().getTocLevel(distance(bgn, start));
 				if (toclevel == Layout::NOT_IN_TOC)
 					continue;
 				DocumentClass::const_iterator lit = tc.begin();
@@ -412,7 +432,7 @@ static void outline(OutlineOp mode, Cursor & cur)
 			pit_type const len = distance(start, finish);
 			buf.undo().recordUndo(cur, ATOMIC_UNDO, pit, pit + len - 1);
 			for (; start != finish; ++start) {
-				toclevel = start->layout().toclevel;
+				toclevel = buf.text().getTocLevel(distance(bgn, start));
 				if (toclevel == Layout::NOT_IN_TOC)
 					continue;
 				DocumentClass::const_iterator lit = tc.begin();
@@ -469,7 +489,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	// at the end?
 	cur.noScreenUpdate();
 
-	LASSERT(cur.text() == this, /**/);
+	LBUFERR(this == cur.text());
 	CursorSlice const oldTopSlice = cur.top();
 	bool const oldBoundary = cur.boundary();
 	bool const oldSelection = cur.selection();
@@ -482,7 +502,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		LyXAction::NoUpdate) || singleParUpdate);
 	bool const last_misspelled = lyxrc.spellcheck_continuously
 		&& cur.paragraph().isMisspelled(cur.pos(), true);
-	
+
 	FuncCode const act = cmd.action();
 	switch (act) {
 
@@ -548,7 +568,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		finishChange(cur, false);
 		break;
 
-	case LFUN_LINE_DELETE:
+	case LFUN_LINE_DELETE_FORWARD:
 		if (cur.selection())
 			cutSelection(cur, true, false);
 		else
@@ -603,7 +623,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 			needsUpdate |= cursorTop(cur);
 			needsUpdate |= cur.selHandle(true);
 			needsUpdate |= cursorBottom(cur);
-		} else 
+		} else
 			cur.undispatched();
 		cur.screenUpdateFlags(Update::FitCursor);
 		break;
@@ -618,13 +638,13 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 				&& cur.boundary() == oldBoundary) {
 			cur.undispatched();
 			cmd = FuncRequest(LFUN_FINISHED_FORWARD);
-		
+
 			// we will probably be moving out the inset, so we should execute
-			// the depm-mechanism, but only when the cursor has a place to 
+			// the depm-mechanism, but only when the cursor has a place to
 			// go outside this inset, i.e. in a slice above.
-			if (cur.depth() > 1 && cur.pos() == cur.lastpos() 
+			if (cur.depth() > 1 && cur.pos() == cur.lastpos()
 				  && cur.pit() == cur.lastpit()) {
-				// The cursor hasn't changed yet. To give the 
+				// The cursor hasn't changed yet. To give the
 				// DEPM the possibility of doing something we must
 				// provide it with two different cursors.
 				Cursor dummy = cur;
@@ -647,10 +667,10 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 			cmd = FuncRequest(LFUN_FINISHED_BACKWARD);
 
 			// we will probably be moving out the inset, so we should execute
-			// the depm-mechanism, but only when the cursor has a place to 
+			// the depm-mechanism, but only when the cursor has a place to
 			// go outside this inset, i.e. in a slice above.
 			if (cur.depth() > 1 && cur.pos() == 0 && cur.pit() == 0) {
-				// The cursor hasn't changed yet. To give the 
+				// The cursor hasn't changed yet. To give the
 				// DEPM the possibility of doing something we must
 				// provide it with two different cursors.
 				Cursor dummy = cur;
@@ -722,15 +742,14 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		bool const atFirstOrLastRow = cur.atFirstOrLastRow(up);
 
 		if (!atFirstOrLastRow) {
-			needsUpdate |= cur.selHandle(select);	
-			cur.selHandle(select);
+			needsUpdate |= cur.selHandle(select);
 			cur.upDownInText(up, needsUpdate);
 			needsUpdate |= cur.beforeDispatchCursor().inMathed();
 		} else {
 			// if the cursor cannot be moved up or down do not remove
 			// the selection right now, but wait for the next dispatch.
 			if (select)
-				needsUpdate |= cur.selHandle(select);	
+				needsUpdate |= cur.selHandle(select);
 			cur.upDownInText(up, needsUpdate);
 			cur.undispatched();
 		}
@@ -773,7 +792,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		ParagraphList::iterator finish = start;
 		ParagraphList::iterator end = pars.end();
 
-		int const thistoclevel = start->layout().toclevel;
+		int const thistoclevel = buf.text().getTocLevel(distance(bgn, start));
 		if (thistoclevel == Layout::NOT_IN_TOC)
 			break;
 
@@ -787,12 +806,12 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 
 		// Seek the one (on same level) below
 		for (; finish != end; ++finish, ++cur.pit()) {
-			int const toclevel = finish->layout().toclevel;
+			int const toclevel = buf.text().getTocLevel(distance(bgn, finish));
 			if (toclevel != Layout::NOT_IN_TOC && toclevel <= thistoclevel)
 				break;
 		}
 		cur.pos() = cur.lastpos();
-		
+
 		needsUpdate |= cur != old_cur;
 		break;
 	}
@@ -829,13 +848,13 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 				&& cur.boundary() == oldBoundary) {
 			cur.undispatched();
 			cmd = FuncRequest(LFUN_FINISHED_FORWARD);
-		
+
 			// we will probably be moving out the inset, so we should execute
-			// the depm-mechanism, but only when the cursor has a place to 
+			// the depm-mechanism, but only when the cursor has a place to
 			// go outside this inset, i.e. in a slice above.
-			if (cur.depth() > 1 && cur.pos() == cur.lastpos() 
+			if (cur.depth() > 1 && cur.pos() == cur.lastpos()
 				  && cur.pit() == cur.lastpit()) {
-				// The cursor hasn't changed yet. To give the 
+				// The cursor hasn't changed yet. To give the
 				// DEPM the possibility of doing something we must
 				// provide it with two different cursors.
 				Cursor dummy = cur;
@@ -873,18 +892,18 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	case LFUN_WORD_BACKWARD_SELECT:
 		needsUpdate |= cur.selHandle(cmd.action() == LFUN_WORD_BACKWARD_SELECT);
 		needsUpdate |= cursorBackwardOneWord(cur);
-	
+
 		if (!needsUpdate && oldTopSlice == cur.top()
 				&& cur.boundary() == oldBoundary) {
 			cur.undispatched();
 			cmd = FuncRequest(LFUN_FINISHED_BACKWARD);
-		
+
 			// we will probably be moving out the inset, so we should execute
-			// the depm-mechanism, but only when the cursor has a place to 
+			// the depm-mechanism, but only when the cursor has a place to
 			// go outside this inset, i.e. in a slice above.
-			if (cur.depth() > 1 && cur.pos() == 0 
+			if (cur.depth() > 1 && cur.pos() == 0
 				  && cur.pit() == 0) {
-				// The cursor hasn't changed yet. To give the 
+				// The cursor hasn't changed yet. To give the
 				// DEPM the possibility of doing something we must
 				// provide it with two different cursors.
 				Cursor dummy = cur;
@@ -905,15 +924,12 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	case LFUN_NEWLINE_INSERT: {
 		InsetNewlineParams inp;
 		docstring arg = cmd.argument();
-		// this avoids a double undo
-		// FIXME: should not be needed, ideally
-		if (!cur.selection())
-			cur.recordUndo();
-		cap::replaceSelection(cur);
 		if (arg == "linebreak")
 			inp.kind = InsetNewlineParams::LINEBREAK;
 		else
 			inp.kind = InsetNewlineParams::NEWLINE;
+		cap::replaceSelection(cur);
+		cur.recordUndo();
 		cur.insert(new InsetNewline(inp));
 		cur.posForward();
 		moveCursor(cur, false);
@@ -929,7 +945,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 			cur.recordUndoSelection();
 			pit_type const pit_end = cur.selEnd().pit();
 			for (pit_type pit = cur.selBegin().pit(); pit <= pit_end; pit++) {
-				pars_[pit].insertChar(0, '\t', 
+				pars_[pit].insertChar(0, '\t',
 						      bv->buffer().params().trackChanges);
 				// Update the selection pos to make sure the selection does not
 				// change as the inserted tab will increase the logical pos.
@@ -943,7 +959,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 			// Maybe we shouldn't allow tabs within a line, because they
 			// are not (yet) aligned as one might do expect.
 			FuncRequest cmd(LFUN_SELF_INSERT, from_ascii("\t"));
-			dispatch(cur, cmd);	
+			dispatch(cur, cmd);
 		}
 		break;
 	}
@@ -963,11 +979,11 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 				if (c == '\t' || c == ' ') {
 					// remove either 1 tab or 4 spaces.
 					int const n = (c == ' ' ? 4 : 1);
-					for (int i = 0; i < n 
+					for (int i = 0; i < n
 						  && !par.empty() && par.getChar(0) == c; ++i) {
 						if (cur.pit() == pit)
 							cur.posBackward();
-						if (cur.realAnchor().pit() == pit 
+						if (cur.realAnchor().pit() == pit
 							  && cur.realAnchor().pos() > 0 )
 							cur.realAnchor().backwardPos();
 						par.eraseChar(0, tc);
@@ -976,23 +992,23 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 			}
 			cur.finishUndo();
 		} else {
-			// If there is no selection, try to remove a tab or some spaces 
+			// If there is no selection, try to remove a tab or some spaces
 			// before the position of the cursor.
 			Paragraph & par = paragraphs()[cur.pit()];
 			pos_type const pos = cur.pos();
-			
+
 			if (pos == 0)
 				break;
-			
+
 			char_type const c = par.getChar(pos - 1);
 			cur.recordUndo();
 			if (c == '\t') {
 				cur.posBackward();
 				par.eraseChar(cur.pos(), tc);
 			} else
-				for (int n_spaces = 0; 
+				for (int n_spaces = 0;
 				     cur.pos() > 0
-					     && par.getChar(cur.pos() - 1) == ' ' 
+					     && par.getChar(cur.pos() - 1) == ' '
 					     && n_spaces < 4;
 				     ++n_spaces) {
 					cur.posBackward();
@@ -1036,7 +1052,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		}
 		break;
 
-	case LFUN_BREAK_PARAGRAPH:
+	case LFUN_PARAGRAPH_BREAK:
 		cap::replaceSelection(cur);
 		breakParagraph(cur, cmd.argument() == "inverse");
 		cur.resetAnchor();
@@ -1065,7 +1081,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 			/*
 			Paragraph & par = pars_[cur.pit()];
 			if (inset->lyxCode() == LABEL_CODE
-				&& par.layout().labeltype == LABEL_COUNTER) {
+				&& !par.layout().counter.empty() {
 				// Go to the end of the paragraph
 				// Warning: Because of Change-Tracking, the last
 				// position is 'size()' and not 'size()-1':
@@ -1157,6 +1173,35 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		break;
 	}
 
+	case LFUN_IPAMACRO_INSERT: {
+		string const arg = cmd.getArg(0);
+		if (arg == "deco") {
+			// Open the inset, and move the current selection
+			// inside it.
+			doInsertInset(cur, this, cmd, true, true);
+			cur.posForward();
+			// Some insets are numbered, others are shown in the outline pane so
+			// let's update the labels and the toc backend.
+			cur.forceBufferUpdate();
+			break;
+		}
+		if (arg == "tone-falling")
+			ipaChar(cur, InsetIPAChar::TONE_FALLING);
+		else if (arg == "tone-rising")
+			ipaChar(cur, InsetIPAChar::TONE_RISING);
+		else if (arg == "tone-high-rising")
+			ipaChar(cur, InsetIPAChar::TONE_HIGH_RISING);
+		else if (arg == "tone-low-rising")
+			ipaChar(cur, InsetIPAChar::TONE_LOW_RISING);
+		else if (arg == "tone-high-rising-falling")
+			ipaChar(cur, InsetIPAChar::TONE_HIGH_RISING_FALLING);
+		else if (arg.empty())
+			lyxerr << "LyX function 'ipamacro-insert' needs an argument." << endl;
+		else
+			lyxerr << "Wrong argument for LyX function 'ipamacro-insert'." << endl;
+		break;
+	}
+
 	case LFUN_WORD_UPCASE:
 		changeCase(cur, text_uppercase);
 		break;
@@ -1175,23 +1220,30 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 
 	case LFUN_PASTE: {
 		cur.message(_("Paste"));
-		LASSERT(cur.selBegin().idx() == cur.selEnd().idx(), /**/);
+		LASSERT(cur.selBegin().idx() == cur.selEnd().idx(), break);
 		cap::replaceSelection(cur);
 
 		// without argument?
 		string const arg = to_utf8(cmd.argument());
 		if (arg.empty()) {
+			bool tryGraphics = true;
 			if (theClipboard().isInternal())
 				pasteFromStack(cur, bv->buffer().errorList("Paste"), 0);
-			else if (theClipboard().hasGraphicsContents() 
-				     && !theClipboard().hasTextContents())
+			else if (theClipboard().hasTextContents()) {
+				if (pasteClipboardText(cur, bv->buffer().errorList("Paste"),
+				                       true, Clipboard::AnyTextType))
+					tryGraphics = false;
+			}
+			if (tryGraphics && theClipboard().hasGraphicsContents())
 				pasteClipboardGraphics(cur, bv->buffer().errorList("Paste"));
-			else
-				pasteClipboardText(cur, bv->buffer().errorList("Paste"));
 		} else if (isStrUnsignedInt(arg)) {
 			// we have a numerical argument
 			pasteFromStack(cur, bv->buffer().errorList("Paste"),
 				       convert<unsigned int>(arg));
+		} else if (arg == "html" || arg == "latex") {
+			Clipboard::TextType type = (arg == "html") ?
+				Clipboard::HtmlTextType : Clipboard::LaTeXTextType;
+			pasteClipboardText(cur, bv->buffer().errorList("Paste"), true, type);
 		} else {
 			Clipboard::GraphicsType type = Clipboard::AnyGraphicsType;
 			if (arg == "pdf")
@@ -1206,9 +1258,10 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 				type = Clipboard::EmfGraphicsType;
 			else if (arg == "wmf")
 				type = Clipboard::WmfGraphicsType;
-
 			else
-				LASSERT(false, /**/);
+				// We used to assert, but couldn't the argument come from, say, the
+				// minibuffer and just be mistyped?
+				LYXERR0("Unrecognized graphics type: " << arg);
 
 			pasteClipboardGraphics(cur, bv->buffer().errorList("Paste"), type);
 		}
@@ -1310,17 +1363,75 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		if (change_layout)
 			setLayout(cur, layout);
 
+		Layout::LaTeXArgMap args = tclass[layout].args();
+		Layout::LaTeXArgMap::const_iterator lait = args.begin();
+		Layout::LaTeXArgMap::const_iterator const laend = args.end();
+		for (; lait != laend; ++lait) {
+			Layout::latexarg arg = (*lait).second;
+			if (arg.autoinsert) {
+				FuncRequest cmd(LFUN_ARGUMENT_INSERT, (*lait).first);
+				lyx::dispatch(cmd);
+			}
+		}
+
+		break;
+	}
+
+	case LFUN_ENVIRONMENT_SPLIT: {
+		bool const outer = cmd.argument() == "outer";
+		Paragraph const & para = cur.paragraph();
+		docstring layout = para.layout().name();
+		depth_type split_depth = cur.paragraph().params().depth();
+		if (outer) {
+			// check if we have an environment in our nesting hierarchy
+			pit_type pit = cur.pit();
+			Paragraph cpar = pars_[pit];
+			while (true) {
+				if (pit == 0 || cpar.params().depth() == 0)
+					break;
+				--pit;
+				cpar = pars_[pit];
+				if (cpar.params().depth() < split_depth
+				    && cpar.layout().isEnvironment()) {
+						layout = cpar.layout().name();
+						split_depth = cpar.params().depth();
+				}
+			}
+		}
+		if (cur.pos() > 0)
+			lyx::dispatch(FuncRequest(LFUN_PARAGRAPH_BREAK));
+		if (outer) {
+			while (cur.paragraph().params().depth() > split_depth)
+				lyx::dispatch(FuncRequest(LFUN_DEPTH_DECREMENT));
+		}
+		bool const morecont = cur.lastpos() > cur.pos();
+		// FIXME This hardcoding is bad
+		docstring const sep =
+				cur.buffer()->params().documentClass().hasLayout(from_ascii("Separator"))
+					? from_ascii("Separator") : from_ascii("--Separator--");
+		lyx::dispatch(FuncRequest(LFUN_LAYOUT, sep));
+		lyx::dispatch(FuncRequest(LFUN_PARAGRAPH_BREAK, "inverse"));
+		if (morecont)
+			lyx::dispatch(FuncRequest(LFUN_DOWN));
+		lyx::dispatch(FuncRequest(LFUN_LAYOUT, layout));
+
 		break;
 	}
 
 	case LFUN_CLIPBOARD_PASTE:
-		cur.clearSelection();
+		cap::replaceSelection(cur);
 		pasteClipboardText(cur, bv->buffer().errorList("Paste"),
 			       cmd.argument() == "paragraph");
 		bv->buffer().errors("Paste");
 		break;
 
+	case LFUN_CLIPBOARD_PASTE_SIMPLE:
+		cap::replaceSelection(cur);
+		pasteSimpleText(cur, cmd.argument() == "paragraph");
+		break;
+
 	case LFUN_PRIMARY_SELECTION_PASTE:
+		cap::replaceSelection(cur);
 		pasteString(cur, theSelection().get(),
 			    cmd.argument() == "paragraph");
 		break;
@@ -1329,6 +1440,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		// Copy the selection buffer to the clipboard stack,
 		// because we want it to appear in the "Edit->Paste
 		// recent" menu.
+		cap::replaceSelection(cur);
 		cap::copySelectionToStack();
 		cap::pasteSelection(bv->cursor(), bv->buffer().errorList("Paste"));
 		bv->buffer().errors("Paste");
@@ -1350,25 +1462,25 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	}
 
 	case LFUN_QUOTE_INSERT: {
-		// this avoids a double undo
-		// FIXME: should not be needed, ideally
-		if (!cur.selection())
-			cur.recordUndo();
 		cap::replaceSelection(cur);
+		cur.recordUndo();
 
 		Paragraph const & par = cur.paragraph();
 		pos_type pos = cur.pos();
+		// Ignore deleted text before cursor
+		while (pos > 0 && par.isDeleted(pos - 1))
+			--pos;
 
 		BufferParams const & bufparams = bv->buffer().params();
-		bool const hebrew = 
+		bool const hebrew =
 			par.getFontSettings(bufparams, pos).language()->lang() == "hebrew";
 		bool const allow_inset_quote = !(par.isPassThru() || hebrew);
-		
+
+		string const arg = to_utf8(cmd.argument());
 		if (allow_inset_quote) {
 			char_type c = ' ';
 			if (pos > 0 && (!cur.prevInset() || !cur.prevInset()->isSpace()))
 				c = par.getChar(pos - 1);
-			string const arg = to_utf8(cmd.argument());
 			InsetQuotes::QuoteTimes const quote_type = (arg == "single")
 				? InsetQuotes::SingleQuotes : InsetQuotes::DoubleQuotes;
 			cur.insert(new InsetQuotes(cur.buffer(), c, quote_type));
@@ -1376,8 +1488,9 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		} else {
 			// The cursor might have been invalidated by the replaceSelection.
 			cur.buffer()->changed(true);
-			lyx::dispatch(FuncRequest(LFUN_SELF_INSERT, "\""));
-		}			
+			string const quote_string = (arg == "single") ? "'" : "\"";
+			lyx::dispatch(FuncRequest(LFUN_SELF_INSERT, quote_string));
+		}
 		break;
 	}
 
@@ -1401,7 +1514,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 
 	case LFUN_MOUSE_DOUBLE:
 		if (cmd.button() == mouse_button::button1) {
-			selectWord(cur, WHOLE_WORD_STRICT);
+			selectWord(cur, WHOLE_WORD);
 			bv->cursor() = cur;
 		}
 		break;
@@ -1559,16 +1672,47 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		break;
 	}
 
-	case LFUN_HYPERLINK_INSERT: {
-		InsetCommandParams p(HYPERLINK_CODE);
-		docstring content;
+	case LFUN_HREF_INSERT: {
+		// FIXME If we're actually given an argument, shouldn't
+		// we use it, whether or not we have a selection?
+		docstring content = cmd.argument();
 		if (cur.selection()) {
 			content = cur.selectionAsString(false);
 			cutSelection(cur, true, false);
 		}
-		p["target"] = (cmd.argument().empty()) ?
-			content : cmd.argument();
+
+		InsetCommandParams p(HYPERLINK_CODE);
+		if (!content.empty()){
+			// if it looks like a link, we'll put it as target,
+			// otherwise as name (bug #8792).
+
+			// We can't do:
+			//   regex_match(to_utf8(content), matches, link_re)
+			// because smatch stores pointers to the substrings rather
+			// than making copies of them. And those pointers become
+			// invalid after regex_match returns, since it is then
+			// being given a temporary object. (Thanks to Georg for
+			// figuring that out.)
+			regex const link_re("^([a-z]+):.*");
+			smatch matches;
+			string const c = to_utf8(lowercase(content));
+
+			if (c.substr(0,7) == "mailto:") {
+				p["target"] = content;
+				p["type"] = from_ascii("mailto:");
+			} else if (regex_match(c, matches, link_re)) {
+				p["target"] = content;
+				string protocol = matches.str(1);
+				if (protocol == "file")
+					p["type"] = from_ascii("file:");
+			} else
+				p["name"] = content;
+		}
 		string const data = InsetCommand::params2string(p);
+
+		// we need to have a target. if we already have one, then
+		// that gets used at the default for the name, too, which
+		// is probably what is wanted.
 		if (p["target"].empty()) {
 			bv->showDialog("href", data);
 		} else {
@@ -1616,7 +1760,6 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	case LFUN_CAPTION_INSERT:
 	case LFUN_FOOTNOTE_INSERT:
 	case LFUN_NOTE_INSERT:
-	case LFUN_FLEX_INSERT:
 	case LFUN_BOX_INSERT:
 	case LFUN_BRANCH_INSERT:
 	case LFUN_PHANTOM_INSERT:
@@ -1627,6 +1770,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	case LFUN_INDEX_INSERT:
 	case LFUN_PREVIEW_INSERT:
 	case LFUN_SCRIPT_INSERT:
+	case LFUN_IPA_INSERT:
 		// Open the inset, and move the current selection
 		// inside it.
 		doInsertInset(cur, this, cmd, true, true);
@@ -1636,6 +1780,37 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		cur.forceBufferUpdate();
 		break;
 
+	case LFUN_FLEX_INSERT: {
+		// Open the inset, and move the current selection
+		// inside it.
+		bool const sel = cur.selection();
+		doInsertInset(cur, this, cmd, true, true);
+		// Insert auto-insert arguments
+		bool autoargs = false;
+		Layout::LaTeXArgMap args = cur.inset().getLayout().latexargs();
+		Layout::LaTeXArgMap::const_iterator lait = args.begin();
+		Layout::LaTeXArgMap::const_iterator const laend = args.end();
+		for (; lait != laend; ++lait) {
+			Layout::latexarg arg = (*lait).second;
+			if (arg.autoinsert) {
+				// The cursor might have been invalidated by the replaceSelection.
+				cur.buffer()->changed(true);
+				FuncRequest cmd(LFUN_ARGUMENT_INSERT, (*lait).first);
+				lyx::dispatch(cmd);
+				autoargs = true;
+			}
+		}
+		if (!autoargs) {
+			if (sel)
+				cur.leaveInset(cur.inset());
+			cur.posForward();
+		}
+		// Some insets are numbered, others are shown in the outline pane so
+		// let's update the labels and the toc backend.
+		cur.forceBufferUpdate();
+		break;
+	}
+
 	case LFUN_TABULAR_INSERT:
 		// if there were no arguments, just open the dialog
 		if (doInsertInset(cur, this, cmd, false, true))
@@ -1648,16 +1823,19 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	case LFUN_FLOAT_INSERT:
 	case LFUN_FLOAT_WIDE_INSERT:
 	case LFUN_WRAP_INSERT: {
-		// will some text be moved into the inset?
-		bool content = cur.selection();
+		// will some content be moved into the inset?
+		bool const content = cur.selection();
+		// does the content consist of multiple paragraphs?
+		bool const singlepar = (cur.selBegin().pit() == cur.selEnd().pit());
 
 		doInsertInset(cur, this, cmd, true, true);
 		cur.posForward();
 
-		// If some text is moved into the inset, doInsertInset 
-		// puts the cursor outside the inset. To insert the
-		// caption we put it back into the inset.
-		if (content)
+		// If some single-par content is moved into the inset,
+		// doInsertInset puts the cursor outside the inset.
+		// To insert the caption we put it back into the inset.
+		// FIXME cleanup doInsertInset to avoid such dances!
+		if (content && singlepar)
 			cur.backwardPos();
 
 		ParagraphList & pars = cur.text()->paragraphs();
@@ -1717,7 +1895,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		dispatch(cur, fr);
 		break;
 	}
-	
+
 	case LFUN_NOMENCL_PRINT:
 	case LFUN_NEWPAGE_INSERT:
 		// do nothing fancy
@@ -1793,7 +1971,7 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		cap::replaceSelection(cur);
 		cur.insert(new InsetMathHull(cur.buffer(), hullSimple));
 		checkAndActivateInset(cur, true);
-		LASSERT(cur.inMathed(), /**/);
+		LASSERT(cur.inMathed(), break);
 		cur.dispatch(cmd);
 		break;
 	}
@@ -1861,14 +2039,14 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 		break;
 	}
 
-	case LFUN_FONT_UULINE: {
+	case LFUN_FONT_UNDERUNDERLINE: {
 		Font font(ignore_font, ignore_language);
 		font.fontInfo().setUuline(FONT_TOGGLE);
 		toggleAndShow(cur, this, font);
 		break;
 	}
 
-	case LFUN_FONT_UWAVE: {
+	case LFUN_FONT_UNDERWAVE: {
 		Font font(ignore_font, ignore_language);
 		font.fontInfo().setUwave(FONT_TOGGLE);
 		toggleAndShow(cur, this, font);
@@ -1890,12 +2068,19 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 	}
 
 	case LFUN_LANGUAGE: {
-		Language const * lang = languages.getLanguage(to_utf8(cmd.argument()));
-		if (!lang)
+		string const lang_arg = cmd.getArg(0);
+		bool const reset = (lang_arg.empty() || lang_arg == "reset");
+		Language const * lang =
+			reset ? reset_language
+			      : languages.getLanguage(lang_arg);
+		// we allow reset_language, which is 0, but only if it
+		// was requested via empty or "reset" arg.
+		if (!lang && !reset)
 			break;
+		bool const toggle = (cmd.getArg(1) != "set");
 		selectWordWhenUnderCursor(cur, WHOLE_WORD_STRICT);
 		Font font(ignore_font, lang);
-		toggleAndShow(cur, this, font, false);
+		toggleAndShow(cur, this, font, toggle);
 		break;
 	}
 
@@ -2237,10 +2422,11 @@ void Text::dispatch(Cursor & cur, FuncRequest & cmd)
 bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 			FuncStatus & flag) const
 {
-	LASSERT(cur.text() == this, /**/);
+	LBUFERR(this == cur.text());
 
 	FontInfo const & fontinfo = cur.real_current_font.fontInfo();
 	bool enable = true;
+	bool allow_in_passthru = false;
 	InsetCode code = NO_CODE;
 
 	switch (cmd.action()) {
@@ -2254,6 +2440,9 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 		break;
 
 	case LFUN_APPENDIX:
+		// FIXME We really should not allow this to be put, e.g.,
+		// in a footnote, or in ERT. But it would make sense in a
+		// branch, so I'm not sure what to do.
 		flag.setOnOff(cur.paragraph().params().startOfAppendix());
 		break;
 
@@ -2287,6 +2476,10 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 			code = INDEX_CODE;
 		else if (cmd.argument() == "index_print")
 			code = INDEX_PRINT_CODE;
+		else if (cmd.argument() == "listings")
+			code = LISTINGS_CODE;
+		else if (cmd.argument() == "mathspace")
+			code = MATH_HULL_CODE;
 		else if (cmd.argument() == "nomenclature")
 			code = NOMENCL_CODE;
 		else if (cmd.argument() == "nomencl_print")
@@ -2309,8 +2502,6 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 			code = VSPACE_CODE;
 		else if (cmd.argument() == "wrap")
 			code = WRAP_CODE;
-		else if (cmd.argument() == "listings")
-			code = LISTINGS_CODE;
 		break;
 
 	case LFUN_ERT_INSERT:
@@ -2386,11 +2577,20 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 		}
 		break;
 	}
-	case LFUN_CAPTION_INSERT:
+	case LFUN_CAPTION_INSERT: {
 		code = CAPTION_CODE;
-		// not allowed in description items
-		enable = !inDescriptionItem(cur);
+		string arg = cmd.getArg(0);
+		bool varia = arg != "LongTableNoNumber";
+		if (cur.depth() > 0) {
+			if (&cur[cur.depth() - 1].inset())
+				varia = cur[cur.depth() - 1].inset().allowsCaptionVariation(arg);
+		}
+		// not allowed in description items,
+		// and in specific insets
+		enable = !inDescriptionItem(cur)
+			&& (varia || arg.empty() || arg == "Standard");
 		break;
+	}
 	case LFUN_NOTE_INSERT:
 		code = NOTE_CODE;
 		// in commands (sections etc.) and description items,
@@ -2420,6 +2620,9 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 		    && cur.buffer()->params().branchlist().empty())
 			enable = false;
 		break;
+	case LFUN_IPA_INSERT:
+		code = IPA_CODE;
+		break;
 	case LFUN_PHANTOM_INSERT:
 		code = PHANTOM_CODE;
 		break;
@@ -2431,9 +2634,55 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 		break;
 	case LFUN_ARGUMENT_INSERT: {
 		code = ARG_CODE;
+		allow_in_passthru = true;
+		string const arg = cmd.getArg(0);
+		if (arg.empty()) {
+			enable = false;
+			break;
+		}
 		Layout const & lay = cur.paragraph().layout();
-		int const numargs = lay.reqargs + lay.optargs;
-		enable = cur.paragraph().insetList().count(ARG_CODE) < numargs;
+		Layout::LaTeXArgMap args = lay.args();
+		Layout::LaTeXArgMap::const_iterator const lait =
+				args.find(arg);
+		if (lait != args.end()) {
+			enable = true;
+			pit_type pit = cur.pit();
+			pit_type lastpit = cur.pit();
+			if (lay.isEnvironment() && !prefixIs(arg, "item:")) {
+				// In a sequence of "merged" environment layouts, we only allow
+				// non-item arguments once.
+				lastpit = cur.lastpit();
+				// get the first paragraph in sequence with this layout
+				depth_type const current_depth = cur.paragraph().params().depth();
+				while (true) {
+					if (pit == 0)
+						break;
+					Paragraph cpar = pars_[pit - 1];
+					if (cpar.layout() == lay && cpar.params().depth() == current_depth)
+						--pit;
+					else
+						break;
+				}
+			}
+			for (; pit <= lastpit; ++pit) {
+				if (pars_[pit].layout() != lay)
+					break;
+				InsetList::const_iterator it = pars_[pit].insetList().begin();
+				InsetList::const_iterator end = pars_[pit].insetList().end();
+				for (; it != end; ++it) {
+					if (it->inset->lyxCode() == ARG_CODE) {
+						InsetArgument const * ins =
+							static_cast<InsetArgument const *>(it->inset);
+						if (ins->name() == arg) {
+							// we have this already
+							enable = false;
+							break;
+						}
+					}
+				}
+			}
+		} else
+			enable = false;
 		break;
 	}
 	case LFUN_INDEX_INSERT:
@@ -2456,13 +2705,21 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 		// not allowed in description items
 		enable = !inDescriptionItem(cur);
 		break;
-	case LFUN_HYPERLINK_INSERT:
+	case LFUN_HREF_INSERT:
 		if (cur.selIsMultiCell() || cur.selIsMultiLine()) {
 			enable = false;
 			break;
 		}
 		code = HYPERLINK_CODE;
 		break;
+	case LFUN_IPAMACRO_INSERT: {
+		string const arg = cmd.getArg(0);
+		if (arg == "deco")
+			code = IPADECO_CODE;
+		else
+			code = IPACHAR_CODE;
+		break;
+	}
 	case LFUN_QUOTE_INSERT:
 		// always allow this, since we will inset a raw quote
 		// if an inset is not allowed.
@@ -2567,6 +2824,20 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 			break;
 		}
 
+		// explicit text type?
+		if (arg == "html") {
+			// Do not enable for PlainTextType, since some tidying in the
+			// frontend is needed for HTML, which is too unsafe for plain text.
+			enable = theClipboard().hasTextContents(Clipboard::HtmlTextType);
+			break;
+		} else if (arg == "latex") {
+			// LaTeX is usually not available on the clipboard with
+			// the correct MIME type, but in plain text.
+			enable = theClipboard().hasTextContents(Clipboard::PlainTextType) ||
+			         theClipboard().hasTextContents(Clipboard::LaTeXTextType);
+			break;
+		}
+
 		// explicit graphics type?
 		Clipboard::GraphicsType type = Clipboard::AnyGraphicsType;
 		if ((arg == "pdf" && (type = Clipboard::PdfGraphicsType))
@@ -2585,6 +2856,7 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 	 }
 
 	case LFUN_CLIPBOARD_PASTE:
+	case LFUN_CLIPBOARD_PASTE_SIMPLE:
 		enable = !theClipboard().empty();
 		break;
 
@@ -2625,12 +2897,13 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 	case LFUN_OUTLINE_OUT:
 		// FIXME: LyX is not ready for outlining within inset.
 		enable = isMainText()
-			&& cur.paragraph().layout().toclevel != Layout::NOT_IN_TOC;
+			&& cur.buffer()->text().getTocLevel(cur.pit()) != Layout::NOT_IN_TOC;
 		break;
 
 	case LFUN_NEWLINE_INSERT:
 		// LaTeX restrictions (labels or empty par)
-		enable = (cur.pos() > cur.paragraph().beginOfBody());
+		enable = !cur.paragraph().isPassThru()
+			&& cur.pos() > cur.paragraph().beginOfBody();
 		break;
 
 	case LFUN_TAB_INSERT:
@@ -2662,13 +2935,13 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 
 	case LFUN_LANGUAGE:
 		enable = !cur.paragraph().isPassThru();
-		flag.setOnOff(to_utf8(cmd.argument()) == cur.real_current_font.language()->lang());
+		flag.setOnOff(cmd.getArg(0) == cur.real_current_font.language()->lang());
 		break;
 
-	case LFUN_BREAK_PARAGRAPH:
+	case LFUN_PARAGRAPH_BREAK:
 		enable = cur.inset().getLayout().isMultiPar();
 		break;
-	
+
 	case LFUN_SPELLING_ADD:
 	case LFUN_SPELLING_IGNORE:
 	case LFUN_SPELLING_REMOVE:
@@ -2678,7 +2951,39 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 	case LFUN_LAYOUT:
 		enable = !cur.inset().forcePlainLayout();
 		break;
-		
+
+	case LFUN_ENVIRONMENT_SPLIT: {
+		// FIXME This hardcoding is bad
+		if (!cur.buffer()->params().documentClass().hasLayout(from_ascii("Separator"))
+		    && !cur.buffer()->params().documentClass().hasLayout(from_ascii("--Separator--"))) {
+			enable = false;
+			break;
+		}
+		if (cmd.argument() == "outer") {
+			// check if we have an environment in our nesting hierarchy
+			bool res = false;
+			depth_type const current_depth = cur.paragraph().params().depth();
+			pit_type pit = cur.pit();
+			Paragraph cpar = pars_[pit];
+			while (true) {
+				if (pit == 0 || cpar.params().depth() == 0)
+					break;
+				--pit;
+				cpar = pars_[pit];
+				if (cpar.params().depth() < current_depth)
+					res = cpar.layout().isEnvironment();
+			}
+			enable = res;
+			break;
+		}
+		else if (cur.paragraph().layout().isEnvironment()) {
+			enable = true;
+			break;
+		}
+		enable = false;
+		break;
+	}
+
 	case LFUN_LAYOUT_PARAGRAPH:
 	case LFUN_PARAGRAPH_PARAMS:
 	case LFUN_PARAGRAPH_PARAMS_APPLY:
@@ -2709,8 +3014,8 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 	case LFUN_FONT_STATE:
 	case LFUN_FONT_UNDERLINE:
 	case LFUN_FONT_STRIKEOUT:
-	case LFUN_FONT_UULINE:
-	case LFUN_FONT_UWAVE:
+	case LFUN_FONT_UNDERUNDERLINE:
+	case LFUN_FONT_UNDERWAVE:
 	case LFUN_TEXTSTYLE_APPLY:
 	case LFUN_TEXTSTYLE_UPDATE:
 		enable = !cur.paragraph().isPassThru();
@@ -2718,7 +3023,7 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 
 	case LFUN_WORD_DELETE_FORWARD:
 	case LFUN_WORD_DELETE_BACKWARD:
-	case LFUN_LINE_DELETE:
+	case LFUN_LINE_DELETE_FORWARD:
 	case LFUN_WORD_FORWARD:
 	case LFUN_WORD_BACKWARD:
 	case LFUN_WORD_RIGHT:
@@ -2793,9 +3098,9 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 	}
 
 	if (code != NO_CODE
-	    && (cur.empty() 
+	    && (cur.empty()
 		|| !cur.inset().insetAllowed(code)
-		|| cur.paragraph().layout().pass_thru))
+		|| (cur.paragraph().layout().pass_thru && !allow_in_passthru)))
 		enable = false;
 
 	flag.setEnabled(enable);
@@ -2806,7 +3111,6 @@ bool Text::getStatus(Cursor & cur, FuncRequest const & cmd,
 void Text::pasteString(Cursor & cur, docstring const & clip,
 		bool asParagraphs)
 {
-	cur.clearSelection();
 	if (!clip.empty()) {
 		cur.recordUndo();
 		if (asParagraphs)
