@@ -80,7 +80,7 @@ InsetCommand::~InsetCommand()
 
 void InsetCommand::metrics(MetricsInfo & mi, Dimension & dim) const
 {
-	button_.update(screenLabel(), editable() || hasSettings());
+	button_.update(screenLabel(), editable() || clickable(0, 0));
 	button_.metrics(mi, dim);
 }
 
@@ -126,7 +126,8 @@ void InsetCommand::latex(otexstream & os, OutputParams const & runparams_in) con
 }
 
 
-int InsetCommand::plaintext(odocstream & os, OutputParams const &) const
+int InsetCommand::plaintext(odocstringstream & os,
+        OutputParams const &, size_t) const
 {
 	docstring const str = "[" + buffer().B_("LaTeX Command: ")
 		+ from_utf8(getCmdName()) + "]";
@@ -210,9 +211,9 @@ bool InsetCommand::getStatus(Cursor & cur, FuncRequest const & cmd,
 }
 
 
-docstring InsetCommand::contextMenuName() const
+string InsetCommand::contextMenuName() const
 {
-	return from_ascii("context-") + from_ascii(insetName(p_.code()));
+	return "context-" + insetName(p_.code());
 }
 
 
@@ -332,6 +333,11 @@ bool decodeInsetParam(string const & name, string & data,
 		data = InsetGraphics::params2string(p, buffer);
 		break;
 	}
+	case MATH_SPACE_CODE: {
+	    InsetSpaceParams p(true);
+		data = InsetSpace::params2string(p);
+		break;
+	}
 	case NOTE_CODE: {
 		InsetNoteParams p;
 		data = InsetNote::params2string(p);
