@@ -4,7 +4,7 @@
  * This file is part of LyX, the document processor.
  * Licence details can be found in the file COPYING.
  *
- * \author Lars Gullik Bjnnes
+ * \author Lars Gullik Bjønnes
  *
  * Full author contact details are available in file CREDITS.
  */
@@ -23,13 +23,12 @@ namespace lyx {
 class InsetFloatList : public InsetCommand {
 public:
 	///
-	InsetFloatList();
+	InsetFloatList(Buffer *);
 	///
-	InsetFloatList(std::string const & type);
-	///
-	docstring screenLabel() const;
-	///
-	EDITABLE editable() const { return IS_EDITABLE; }
+	InsetFloatList(Buffer *, std::string const & type);
+
+	/// \name Public functions inherited from Inset class
+	//@{
 	///
 	InsetCode lyxCode() const { return FLOAT_LIST_CODE; }
 	///
@@ -39,22 +38,47 @@ public:
 	///
 	void read(Lexer &);
 	///
-	int latex(odocstream &, OutputParams const &) const;
+	void latex(otexstream &, OutputParams const &) const;
 	///
 	int docbook(odocstream &, OutputParams const &) const { return 0; }
 	///
-	int plaintext(odocstream &, OutputParams const & runparams) const;
+	int plaintext(odocstringstream & ods, OutputParams const & op,
+	              size_t max_length = INT_MAX) const;
+	///
+	docstring xhtml(XHTMLStream &, OutputParams const &) const;
+	///
+	void doDispatch(Cursor & cur, FuncRequest & cmd);
+	///
+	bool clickable(int, int) const { return true; }
 	///
 	void validate(LaTeXFeatures & features) const;
+	//@}
+
+	/// \name Static public methods obligated for InsetCommand derived classes
+	//@{
 	///
 	static ParamInfo const & findInfo(std::string const &);
 	///
-	static std::string defaultCommand() { return "listoftables"; };
+	static std::string defaultCommand() { return "listoftables"; }
 	///
 	static bool isCompatibleCommand(std::string const & s);
+	//@}
+
 private:
+	/// \name Private functions inherited from Inset class
+	//@{
 	///
 	Inset * clone() const { return new InsetFloatList(*this); }
+	///
+	docstring layoutName() const;
+	//@}
+
+	/// \name Private functions inherited from InsetCommand class
+	//@{
+	///
+	docstring screenLabel() const;
+	//@}
+
 	///
 	static ParamInfo param_info_;
 };
