@@ -23,7 +23,11 @@ public:
 	enum Kind {
 		phantom,
 		vphantom,
-		hphantom
+		hphantom,
+		smash,
+		mathclap,
+		mathllap,
+		mathrlap
 	};
 	///
 	explicit InsetMathPhantom(Buffer * buf, Kind);
@@ -37,6 +41,17 @@ public:
 	void normalize(NormalStream & ns) const;
 	///
 	void infoize(odocstream & os) const;
+	///
+	InsetCode lyxCode() const { return MATH_PHANTOM_CODE; }
+	/// Nothing for now
+	void mathmlize(MathStream &) const {}
+	/// Nothing for HTML
+	void htmlize(HtmlStream &) const {}
+	/// request "external features"
+	void validate(LaTeXFeatures & features) const;
+	/// Does the contents appear in LaTeX output?
+	bool visibleContents() const;
+
 private:
 	///
 	virtual Inset * clone() const;
