@@ -34,6 +34,7 @@
 #include "frontends/Painter.h"
 
 #include "support/debug.h"
+#include "support/gettext.h"
 #include "support/lassert.h"
 #include "support/textutils.h"
 
@@ -125,7 +126,8 @@ private:
 MathMacro::MathMacro(Buffer * buf, docstring const & name)
 	: InsetMathNest(buf, 0), name_(name), displayMode_(DISPLAY_INIT),
 		expanded_(buf), attachedArgsNum_(0), optionals_(0), nextFoldMode_(true),
-		macroBackup_(buf), macro_(0), needsUpdate_(false), appetite_(9)
+		macroBackup_(buf), macro_(0), needsUpdate_(false),
+		isUpdating_(false), appetite_(9)
 {}
 
 
@@ -215,7 +217,7 @@ void MathMacro::metrics(MetricsInfo & mi, Dimension & dim) const
 		   && editing_[mi.base.bv]) {
 		// Macro will be edited in a old-style list mode here:
 
-		LASSERT(macro_ != 0, /**/);
+		LBUFERR(macro_);
 		Dimension fontDim;
 		FontInfo labelFont = sane_font;
 		math_font_max_dim(labelFont, fontDim.asc, fontDim.des);
@@ -252,7 +254,7 @@ void MathMacro::metrics(MetricsInfo & mi, Dimension & dim) const
 		dim.wid += 2;
 		metricsMarkers2(dim);
 	} else {
-		LASSERT(macro_ != 0, /**/);
+		LBUFERR(macro_);
 
 		// calculate metrics, hoping that all cells are seen
 		macro_->lock();
@@ -311,8 +313,27 @@ void MathMacro::updateMacro(MacroContext const & mc)
 }
 
 
-void MathMacro::updateRepresentation()
+class MathMacro::UpdateLocker
 {
+public:
+	explicit UpdateLocker(MathMacro & mm) : mac(mm)
+	{
+		mac.isUpdating_ = true;
+	}
+	~UpdateLocker() { mac.isUpdating_ = false; }
+private:
+	MathMacro & mac;
+};
+
+
+void MathMacro::updateRepresentation(Cursor * cur, MacroContext const & mc,
+		UpdateType utype)
+{
+	if (isUpdating_)
+		return;
+
+	UpdateLocker(*this);
+
 	// known macro?
 	if (macro_ == 0)
 		return;
@@ -342,6 +363,8 @@ void MathMacro::updateRepresentation()
 	}
 	// expanding macro with the values
 	macro_->expand(values, expanded_.cell(0));
+	if (utype == OutputUpdate && !expanded_.cell(0).empty())
+		expanded_.cell(0).updateMacros(cur, mc, utype);
 	// get definition for list edit mode
 	docstring const & display = macro_->display();
 	asArray(display.empty() ? macro_->definition() : display, definition_);
@@ -462,7 +485,7 @@ void MathMacro::draw(PainterInfo & pi, int x, int y) const
 void MathMacro::drawSelection(PainterInfo & pi, int x, int y) const
 {
 	// We may have 0 arguments, but InsetMathNest requires at least one.
-	if (cells_.size() > 0)
+	if (!cells_.empty())
 		InsetMathNest::drawSelection(pi, x, y);
 }
 
@@ -504,8 +527,7 @@ bool MathMacro::validName() const
 {
 	docstring n = name();
 
-	// empty name?
-	if (n.size() == 0)
+	if (n.empty())
 		return false;
 
 	// converting back and force doesn't swallow anything?
@@ -562,7 +584,7 @@ Inset * MathMacro::editXY(Cursor & cur, int x, int y)
 
 void MathMacro::removeArgument(Inset::pos_type pos) {
 	if (displayMode_ == DISPLAY_NORMAL) {
-		LASSERT(size_t(pos) < cells_.size(), /**/);
+		LASSERT(size_t(pos) < cells_.size(), return);
 		cells_.erase(cells_.begin() + pos);
 		if (size_t(pos) < attachedArgsNum_)
 			--attachedArgsNum_;
@@ -577,7 +599,7 @@ void MathMacro::removeArgument(Inset::pos_type pos) {
 
 void MathMacro::insertArgument(Inset::pos_type pos) {
 	if (displayMode_ == DISPLAY_NORMAL) {
-		LASSERT(size_t(pos) <= cells_.size(), /**/);
+		LASSERT(size_t(pos) <= cells_.size(), return);
 		cells_.insert(cells_.begin() + pos, MathData());
 		if (size_t(pos) < attachedArgsNum_)
 			++attachedArgsNum_;
@@ -591,7 +613,7 @@ void MathMacro::insertArgument(Inset::pos_type pos) {
 
 void MathMacro::detachArguments(vector<MathData> & args, bool strip)
 {
-	LASSERT(displayMode_ == DISPLAY_NORMAL, /**/);	
+	LASSERT(displayMode_ == DISPLAY_NORMAL, return);
 	args = cells_;
 
 	// strip off empty cells, but not more than arity-attachedArgsNum_
@@ -612,7 +634,7 @@ void MathMacro::detachArguments(vector<MathData> & args, bool strip)
 
 void MathMacro::attachArguments(vector<MathData> const & args, size_t arity, int optionals)
 {
-	LASSERT(displayMode_ == DISPLAY_NORMAL, /**/);
+	LASSERT(displayMode_ == DISPLAY_NORMAL, return);
 	cells_ = args;
 	attachedArgsNum_ = args.size();
 	cells_.resize(arity);
@@ -645,7 +667,9 @@ bool MathMacro::notifyCursorLeaves(Cursor const & old, Cursor & cur)
 			// The macro name was changed
 			Cursor inset_cursor = old;
 			int macroSlice = inset_cursor.find(this);
-			LASSERT(macroSlice != -1, /**/);
+			// returning true means the cursor is "now" invalid,
+			// which it was.
+			LASSERT(macroSlice != -1, return true);
 			inset_cursor.cutOff(macroSlice);
 			inset_cursor.recordUndoInset();
 			inset_cursor.pop();
@@ -699,10 +723,11 @@ void MathMacro::write(WriteStream & os) const
 	}
 
 	// normal mode
-	LASSERT(macro_, /**/);
+	// we should be ok to continue even if this fails.
+	LATTEST(macro_);
 
-	// optional arguments make macros fragile
-	if (optionals_ > 0 && os.fragile())
+	// Always protect macros in a fragile environment
+	if (os.fragile())
 		os << "\\protect";
 	
 	os << "\\" << name();
@@ -730,7 +755,7 @@ void MathMacro::write(WriteStream & os) const
 	for (; i < cells_.size(); ++i) {
 		if (cell(i).size() == 1 
 			&& cell(i)[0].nucleus()->asCharInset()
-			&& cell(i)[0].nucleus()->asCharInset()->getChar() < 0x80) {
+			&& isASCII(cell(i)[0].nucleus()->asCharInset()->getChar())) {
 			if (first)
 				os << " ";
 			os << cell(i);
